!=========================================================================
! Copyright (C) GemTalk Systems 1986-2017.  All Rights Reserved.
!
! $Id: object2.gs 42264 2017-07-28 05:24:38Z lalmarod $
!
! more methods for Object 
!
! Superclass Hierarchy:
!   Object.
!
!=========================================================================
 
! do not remove methods

set class Object
category: 'Error Handling'

method: 
_error: errorSymbol
  "Will be deprecated."
  | errNum cls |
  errNum := ErrorSymbols at: errorSymbol otherwise: 2063 .
  cls := ( (LegacyErrNumMap atOrNil: errNum) 
             ifNotNil:[ :mv | mv atOrNil: 1 ] ) ifNil:[ InternalError ].
  ^ cls new _number: errNum; reason: errorSymbol ; args: { self } ; signal 
%

category: 'Error Handling'
method: Object
_error: errorSymbol with: argOne
  "Will be deprecated."
  | errNum cls |
  errNum := ErrorSymbols at: errorSymbol otherwise: 2063 .
  cls := ( (LegacyErrNumMap atOrNil: errNum) 
             ifNotNil:[ :mv | mv atOrNil: 1 ] ) ifNil:[ InternalError ].
  ^ cls new _number: errNum; reason: errorSymbol ; 
		args: { self . argOne } ; signal
%

category: 'Error Handling'
method: Object
_error: errorSymbol args: argList
  "Will be deprecated."

  | newArgList cls errNum |
  newArgList := { self } .
  newArgList addAll: argList .
  errNum := ErrorSymbols at: errorSymbol otherwise: 2063 .
  cls := ( (LegacyErrNumMap atOrNil: errNum) 
             ifNotNil:[ :mv | mv atOrNil: 1 ] ) ifNil:[ InternalError ].
  ^ cls new _number: errNum; reason: errorSymbol ;
                args: newArgList ;  signal
%

!  _errorNotOnlyUser: deleted, no senders


category: 'Updating'
classmethod: Object
_checkKindForBecome: anObj

"determine reason for primitive failure of become"
| arr |

anObj isSpecial ifTrue: [ ^ anObj _error: #rtErrCantBecomeSpecial ].

anObj _isExecBlock ifTrue: [ 
   ^ ArgumentTypeError new reason: #rtErrCantBecomeClassKind ; _number: 2121; object: anObj ;
	details: 'become not allowed on ExecBlocks  '; signal
].
(anObj _isSymbol ) ifTrue: [ 
   ^ ArgumentTypeError new reason: #rtErrCantBecomeClassKind ; _number: 2121; object: anObj ;
	details: 'become not allowed on Symbols  '; signal
].

arr := { Behavior . GsNMethod .   " to avoid breaking VM"
   GsMethodDictionary . SymbolDictionary . SymbolList .  "for code security"
   "following have CData" GsSocket . GsFile . Regexp . 
     GsCompilerIRNode . CBuffer . CByteArray . CFunction . CLibrary . CPointer . 
     CZstream }.
1 to: arr size do:[:n || cls |
  cls := arr at: n .
  (anObj isKindOf: cls) ifTrue:[ 
    ^ ArgumentTypeError new reason: #rtErrCantBecomeClassKind; _number: 2121 ; object: anObj;
       details: 'become not allowed on instances of ' , cls name ; signal .
  ]
].
%

! extent0.ruby.dbf overrides implementation of _doesNotUnderstand:args:envId:reason:

category: 'Private'
method: Object
_doesNotUnderstand: aSymbol args: anArray envId: aSmallInt reason: dnuKind
	"This private message is sent by the virtual machine. The implementation must
 be in the persistent method dictionary for Object, since the method is preloaded
 during VM startup.  Reimplementations in a session methods dictionary will not
 be seen."

	dnuKind ~~ 0
		ifTrue: [self _dnuError: aSymbol args: anArray reason: dnuKind].
	aSmallInt == 0
		ifTrue: 
			["this path for compatibility with Gs64 v2.x, and assumes you are using
   only method environment 0  for all of your Smalltalk code."
			^self doesNotUnderstand: (Message selector: aSymbol arguments: anArray)].
	^self doesNotUnderstand: aSymbol args: anArray envId: aSmallInt
%

! fixed 41026
category: 'Error Handling'
method: Object
cantPerform: aSelectorSymbol withArguments: anArray
	"This method implements the default response when a message can't be performed
 with _perform:withArguments:.  It raises the rtErrCantPerform exception."

	aSelectorSymbol numArgs == anArray size
		ifTrue: 
			[^self
				doesNotUnderstand: (Message selector: aSelectorSymbol arguments: anArray)].
	self _error: #rtErrCantPerform
		args: 
			{aSelectorSymbol.
			anArray}.
	^self _primitiveFailed: #cantPerform:withArguments:
		args: 
			{aSelectorSymbol.
			anArray}
%
